{+--------------------------------------------------------------------------+
 | Class:       TmTSearch  ( TurboSearch )
 | Created:     8.97
 | Author:      Martin Waldenburg
 | Description: A very fast search engine, about twice as fast as Boyer-Moore,
 |              based on an article in the German magazine c't (8/97).
 |              This implementation is in many cases faster as the original,
 |              but not suited for general pattern matching.
 |              For example it will find all occurences of "one million" but
 |              not all of 1000000.
 |              This is because of the check for the half pattern length.
 |              You may remove it and the corresponding jump.
 |              The original is in 'C '. www.Heise.com.
 |              You can search case sensitive or case insensitive.
  | Version:     2.2
 | Copyright (c) 1997 - 1998 Martin Waldenburg
 | All rights reserved.
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Software using this code must contain a visible line of credit,
 |    to Martin Waldenburg and the German magazine c't (8/97).
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de

+--------------------------------------------------------------------------+}
unit mwTSearch;


interface

uses
  Windows,
  SysUtils,
  Messages,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  Menus;

type
  TmTSearch = class(Tobject)
  private
    Text, Pat: String;
    fCount: Integer;
    Look_At: Integer;
    SimpleJump: Integer;
    fPos: Integer;
    HalfLen, PatLen, PatLenMinus, PatLenPlus, TextLen: Integer;
    Shift: array[0..255] of Integer;
    fFinished: Boolean;
    fFound: Boolean;
    function GetFinished: Boolean;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Init(NewPattern: String);
    function FindFirst(NewText: String): Integer;
    function Next: Integer;
    procedure IInit(NewPattern: String);
    function IFindFirst(NewText: String): Integer;
    function INext: Integer;

    procedure InitO(NewPattern: String);
    function FindFirstO(NewText: String): Integer;
    function NextO: Integer;
    procedure IInitO(NewPattern: String);
    function IFindFirstO(NewText: String): Integer;
    function INextO: Integer;

    procedure InitSJ(NewPattern: String);
    function FindFirstSJ(NewText: String): Integer;
    function NextSJ: Integer;
    procedure IInitSJ(NewPattern: String);
    function IFindFirstSJ(NewText: String): Integer;
    function INextSJ: Integer;

    property Count: Integer read fCount write fCount;
    property Finished: Boolean read GetFinished;
    property Found: Boolean read fFound;
  published
  end;

var
  CompTable: array[#0..#255] of byte;

implementation

constructor TmTSearch.Create;
begin
  inherited Create;
  Pat := '';
  PatLen := 0;
  HalfLen := 0;
  Text := '';
  TextLen := 0;
  fPos := 0;
  fFound := False;
  fCount := 0;
end; { Create }

procedure MakeCompTable;
var
  I: Byte;
begin
  for I := 0 to 255 do
    CompTable[Char(I)] := ord(AnsiLowerCase(Char(I))[1]);
end; { MakeCompTable }

function TmTSearch.GetFinished: Boolean;
begin
  fFinished := False;
  if fPos >= TextLen then fFinished := True;
  if PatLen > TextLen then fFinished := True;
  Result := fFinished;
end; { GetFinished }

procedure TmTSearch.Init(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  HalfLen := PatLen div 2;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[ord(Pat[I])] := PatLenPlus - I;
end; { Init }

function TmTSearch.FindFirst(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := Next;
  end;
end; { FindFirst }

function TmTSearch.Next: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (Pat[PatLen] <> Text[fPos]) then inc(fPos, Shift[ord(Text[fPos + 1])])
    else
    begin
      J := fPos - 1;
      I := PatLen - 1;
      while Pat[I] = Text[J] do
      begin
        dec(I);
        if I = 0 then
        begin
          fFound := True;
          inc(fCount);
          Result := fPos - Patlen + 1;
          exit;
        end;
        dec(J);
      end;
      Case I < HalfLen of
        True: inc(fPos, PatLenPlus);
        False: inc(fPos, Shift[ord(Text[fPos + 1])]);
      end;
    end;
  end;
end; { Next }

procedure TmTSearch.IInit(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  HalfLen := PatLen div 2;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[CompTable[Pat[i]]] := PatLenPlus - I;
end; { IInit }

function TmTSearch.IFindFirst(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := INext;
  end;
end; { IFindFirst }

function TmTSearch.INext: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (CompTable[Pat[PatLen]] <> CompTable[Text[fPos]]) then
      inc(fPos, Shift[CompTable[Text[fPos + 1]]])
    else
    begin
      J := fPos - 1;
      I := PatLen - 1;
      while CompTable[Pat[I]] = CompTable[Text[J]] do
      begin
        dec(I);
        if I = 0 then
        begin
          fFound := True;
          inc(fCount);
          Result := fPos - Patlen + 1;
          exit;
        end;
        dec(J);
      end;
      Case I < HalfLen of
        True: inc(fPos, PatLenPlus);
        False: inc(fPos, Shift[CompTable[Text[fPos + 1]]]);
      end;
    end;
  end;
end; { INext }







procedure TmTSearch.InitO(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  PatLenMinus := PatLen - 1;
  Look_At := 0;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[ord(Pat[I])] := PatLenPlus - I;
  while Look_at < PatLenMinus do
    if Pat[PatLen] = Pat[PatLen - (Look_at + 2)] then exit
    else inc(Look_at);
end; { Init }

function TmTSearch.FindFirstO(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := NextO;
  end;
end; { FindFirstO }

function TmTSearch.NextO: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (Pat[Patlen] <> Text[fPos]) then inc(fPos, Shift[ord(Text[fPos + 1])])
    else
    begin
      J := fPos;
      I := PatLen;
      repeat
        dec(I); dec(J);
      until (I = 0) or (Pat[I] <> Text[J]);
      Case I = 0 of
        True:
          begin
            fFound := True;
            inc(fCount);
            Result := fPos - Patlen + 1;
            exit;
          end;
        False: inc(fPos, Look_at + Shift[ord(Text[fPos + Look_at + 1])]);
      end;
    end;
  end;
end; { NextO }

procedure TmTSearch.IInitO(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  PatLenMinus := PatLen - 1;
  Look_At := 0;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[CompTable[Pat[i]]] := PatLenPlus - I;
  while Look_at < PatLenMinus do
    if CompTable[Pat[PatLen]] = CompTable[Pat[PatLen - (Look_at + 2)]] then exit
    else inc(Look_at);
end; { IInitO }

function TmTSearch.IFindFirstO(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := INextO;
  end;
end; { IFindFirstO }

function TmTSearch.INextO: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (CompTable[Pat[Patlen]] <> CompTable[Text[fPos]]) then
      inc(fPos, Shift[CompTable[Text[fPos + 1]]])
    else
    begin
      J := fPos;
      I := PatLen;
      repeat
        dec(I); dec(J);
      until (I = 0) or (CompTable[Pat[I]] <> CompTable[Text[J]]);
      Case I = 0 of
        True:
          begin
            fFound := True;
            inc(fCount);
            Result := fPos - Patlen + 1;
            exit;
          end;
        False: inc(fPos, Look_At + Shift[CompTable[Text[fPos + Look_at + 1]]]);
      end;
    end;
  end;
end; { INextO }










procedure TmTSearch.InitSJ(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  PatLenMinus := PatLen - 1;
  SimpleJump := 1;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[ord(Pat[I])] := PatLenPlus - I;
  while SimpleJump < PatLen + 2 do
    if Pat[PatLen] = Pat[PatLen - (SimpleJump + 2)] then exit
    else inc(SimpleJump);
end; { InitSJ }

function TmTSearch.FindFirstSJ(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := NextSJ;
  end;
end; { FindFirstSJ }

function TmTSearch.NextSJ: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (Pat[PatLen] <> Text[fPos]) then inc(fPos, Shift[ord(Text[fPos + 1])])
    else
    begin
      J := fPos;
      I := PatLen;
      repeat
        dec(I); dec(J);
      until (I = 0) or (Pat[I] <> Text[J]);
      Case I = 0 of
        True:
          begin
            fFound := True;
            inc(fCount);
            Result := fPos - Patlen + 1;
            exit;
          end;
        False: inc(fPos, SimpleJump);
      end;
    end;
  end;
end; { NextSJ }

procedure TmTSearch.IInitSJ(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  PatLenMinus := PatLen - 1;
  SimpleJump := 1;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[CompTable[Pat[i]]] := PatLenPlus - I;
  while SimpleJump < PatLen + 2 do
    if CompTable[Pat[PatLen]] = CompTable[Pat[PatLen - (SimpleJump + 2)]] then exit
    else inc(SimpleJump);
end; { IInitSJ }

function TmTSearch.IFindFirstSJ(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := INextSJ;
  end;
end; { IFindFirstSJ }

function TmTSearch.INextSJ: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (CompTable[Pat[PatLen]] <> CompTable[Text[fPos]]) then
      inc(fPos, Shift[CompTable[Text[fPos + 1]]])
    else
    begin
      J := fPos;
      I := PatLen;
      repeat
        dec(I); dec(J);
      until (I = 0) or (CompTable[Pat[I]] <> CompTable[Text[J]]);
      Case I = 0 of
        True:
          begin
            fFound := True;
            inc(fCount);
            Result := fPos - Patlen + 1;
            exit;
          end;
        False: inc(fPos, SimpleJump);
      end;
    end;
  end;
end; { INextSJ }


destructor TmTSearch.Destroy;
begin
  inherited Destroy;
end; { Destroy }

Initialization
  MakeCompTable;

end.

